
// Duplicate files linker
//
// Copyright (c)2003 by reanimatolog@yandex.ru, http://bootcd.narod.ru
//
// Independent implementation of MD5 (RFC 1321) by L. Peter Deutsch, ghost@aladdin.com
// (see md5.h/md5.c for more details)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <sys\stat.h>
#include "md5.h"

#define DFL_VERSION "v0.1a"

struct dirinfo{
  char *name;
  int level;
  };

struct fileinfo{
  char *name;
  md5_byte_t *md5digest;
  struct stat lstat;
  struct fileinfo *prev;
  struct fileinfo *next;
  struct fileinfo *linkto;
  struct dirinfo *path;
  int backup;
  };
 
struct fileinfo *firstfile=NULL;
long totalfiles=0;
long long totalbytes=0;
long totalduplicatefiles=0;
long long totalduplicatebytes=0;
int mode=0; // 0 - scan, 1 - link, 2 - remove
int verbose=0;
int disableundo=0;

int main(int argc,char *argv[]){

  char cwdsave[256];

  int status=0;

#if !defined(__DJGPP__) && !defined(__CYGWIN32__) && !defined(__CYGWIN__)
  printf("Duplicate files linker %s for mkisofs\n"
#else
#ifdef __DJGPP__
  printf("Duplicate files linker %s for MS-DOS version of mkisofs\n"
#else
  printf("Duplicate files linker %s for Win32 version of mkisofs\n"
#endif
#endif
  	 "Copyright (c)2003 by reanimatolog@yandex.ru, http://bootcd.narod.ru\n"
         "Independent implementation of MD5 (RFC 1321) by L. Peter Deutsch, ghost@aladdin.com\n"
         "\n"
         "This program is distributed in the hope that it will be useful,\n"
         "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
         "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
         "GNU General Public License for more details.\n"
         "\n",DFL_VERSION);
  if(argc<2){
    status=1;
    }
  else{
    if(argc>2){
      int i;
      for(i=2;i<argc;i++){
        if(!stricmp(argv[i],"find")){
          mode=0;
          }
        else if(!stricmp(argv[i],"link")){
          mode=1;
          }
        else if(!stricmp(argv[i],"remove")){
          mode=2;
          }
        else if(!stricmp(argv[i],"-v")){
          verbose=1;
          }
        else if(!stricmp(argv[i],"-d")){
          disableundo=1;
          }
        else{
          status=1;
          }
        }
      }
    }
  if(!status){
    if(getcwd(cwdsave,256)){
      if(chdir(argv[1])){
        status=2;
        }
      }
    else{
      status=3;
      }
    }
  if(!status){
    if(scan_tree("",0)){
      status=3;
      }
    }
  if(!status){
    if(calculate_md5_digests()){
      status=3;
      }
    }
  if(!status){
    if(find_duplicates()){
      status=3;
      }
    }
  if(!status && totalduplicatefiles && mode){
    if(!create_links()){
      if(!disableundo){
        if(remove_backups()){
          status=4;
          }
        }
      }
    else{
      status=4;
      if(!disableundo){
        if(!undo()){
          status=3;
          }
        }
      }
    }
  switch(status){
    case 0: printf("Ok.\n");
            break;
    case 1: printf("Usage: %s <path> [command] [options]\n"
                   "\n"
                   "Commands:\n"
                   "  find - Find duplicate files (default)\n"
#if !defined(__DJGPP__) && !defined(__CYGWIN32__) && !defined(__CYGWIN__)
                   "  link - Replace duplicate files to symbolic links\n"
#else
#ifdef __DJGPP__
                   "  link - Replace duplicate files to DJGPP's symbolic links\n"
#else
                   "  link - Replace duplicate files to CyrWin's symbolic links\n"
#endif
#endif
                   "remove - Remove duplicate files\n"
                   "\n"
                   "Options:\n"
                   "    -d - Disable undo on errors\n"
                   "    -v - Verbose\n"
                   "\n"
                   "Errorlevels:\n"
                   "     0 - Ok\n"
                   "     1 - Invalid arguments\n"
                   "     2 - Path not found\n"
                   "     3 - Errors found, but tree not changed\n"
                   "     3 - Errors found, tree corrupted\n"
                   "\n"
                   "Example:\n"
                   "     dfl cdfiles link -d\n"
                   "     mkisofs -follow-links -cache-inodes -output cdfiles.iso cdfiles\n",
                   argv[0]);
            break;
    case 2: printf("Path not found.\n");
            break;
    case 3: printf("Errors found, but tree not changed.\n");
            break;
    case 4: printf("Errors found, tree corrupted.\n");
            break;
    }
  chdir(cwdsave);
  exit(status);
  }

// All functions: 0 - Ok, 1 - Not enough memory, 2 - other errors
  
int scan_tree(char *currdirname,int currlevel){

  int status=0;
  struct dirinfo *currdirinfo;
  DIR *currdir=NULL;
  struct dirent	*direntry;
  char *direntryname;
  struct stat direntrylstat;
  struct fileinfo *nextfile;

  if(currlevel){
    if(!(currdir=opendir(currdirname))){
      printf("Can't open directory: %s\n",currdirname);
      status=2;
      }
    else{
      if(verbose){
        printf("/%s\n",currdirname);
        }
      }
    }
  else{
    printf("Scanning tree...\n");
    if(!(currdir=opendir("."))){
      status=2;
      }
    else{
      if(verbose){
        printf("/\n");
        }
      }
    }
  if(!status){ 
    if(!(currdirinfo=(struct dirinfo *)malloc(sizeof(struct dirinfo)))){
      status=1;
      }
    else{
      currdirinfo->name=currdirname;
      currdirinfo->level=currlevel;
      }
    }
  if(!status){
    while(direntry=readdir(currdir)){
      if(!(direntryname=(char *)malloc(strlen(currdirname)+strlen(direntry->d_name)+1+1))){
        status=1;
        break;
        }
      strcpy(direntryname,currdirname);
      strcat(direntryname,direntry->d_name);
      if(lstat(direntryname,&direntrylstat)){
        printf("Can't stat file or directory: %s\n",direntryname);
        status=2;
        break;
        }
      if(!S_ISLNK(direntrylstat.st_mode)){
        if(strcmp(direntry->d_name,".") && strcmp(direntry->d_name,"..")){
          if(S_ISDIR(direntrylstat.st_mode)){
            strcat(direntryname,"/");
            if(status=scan_tree(direntryname,currlevel+1)){
              break;
              }
            continue;
            }
          else{
            totalfiles++;
            totalbytes+=direntrylstat.st_size;
            if(!(nextfile=(struct fileinfo *)malloc(sizeof(struct fileinfo)))){
              status=1;
              break;
              }
            nextfile->md5digest=NULL;
            nextfile->lstat=direntrylstat;
            if(!(nextfile->name=(char *)malloc(strlen(direntry->d_name)+1))){
              status=1;
              break;
              }
            strcpy(nextfile->name,direntry->d_name);
            nextfile->prev=nextfile;
            nextfile->next=nextfile;
            nextfile->path=currdirinfo;
            nextfile->linkto=NULL;
            nextfile->backup=0;
            if(!firstfile){
              firstfile=nextfile;
              }
            else{
              struct fileinfo *tmp=firstfile;
              while(1){
                if(tmp->lstat.st_size > nextfile->lstat.st_size){
                  tmp->prev->next=nextfile;
                  nextfile->next=tmp;
                  nextfile->prev=tmp->prev;
                  tmp->prev=nextfile;
                  if(tmp==firstfile){
                    firstfile=nextfile;
                    }
                  break;
                  }
                tmp=tmp->next;
                if(tmp==firstfile){
                  tmp->prev->next=nextfile;
                  nextfile->next=tmp;
                  nextfile->prev=tmp->prev;
                  tmp->prev=nextfile;
                  break;
                  }
                }
              }
            }
          }
        }
      free(direntryname);
      }
    }
  if(currdir){
    closedir(currdir);
    }
  if(status==1 && !currlevel){
    printf("Not enough memory.\n");
    }
  if(!status && !currlevel){
    printf("Total %llu byte(s) in %lu file(s)\n",totalbytes,totalfiles);
    }
  return status;
  }

int calculate_md5_digests(){

  int status=0;
  struct fileinfo *tmp=firstfile;
  char *filename=NULL;
  FILE *md5file=NULL;
  md5_state_t md5state;
  off_t md5count;
  off_t md5portion;
  char md5buf[65536];
  char md5text[16*2+1]={0};

  if(tmp){
    printf("Calculating MD5 digests...\n");
    do{
      if(tmp->lstat.st_size && (tmp->lstat.st_size == tmp->next->lstat.st_size || tmp->lstat.st_size == tmp->prev->lstat.st_size)){
        if(!(filename=(char *)malloc(strlen(tmp->path->name)+strlen(tmp->name)+1))){
          status=1;
          break;
          }
        strcpy(filename,tmp->path->name);
        strcat(filename,tmp->name);
        if((md5file=fopen(filename,"rb"))==NULL){
          printf("Can't open file: %s\n",filename);
          status=2;
          break;
          }
        md5count=tmp->lstat.st_size;
        md5_init(&md5state);
        while(md5count){
          md5portion=(md5count>65536)?65536:md5count;
          if(fread(md5buf,md5portion,1,md5file)!=1){
            printf("Can't read file: %s\n",filename);
            status=2;
            break;
            }
          md5_append(&md5state,(const md5_byte_t *)md5buf,md5portion);
          md5count-=md5portion;
          }
        fclose(md5file);
        free(filename);
        if(md5count){
          break;
          }
        if(!(tmp->md5digest=(md5_byte_t *)malloc(sizeof(md5_byte_t)*16))){
          status=1;
          break;
          }
        md5_finish(&md5state,tmp->md5digest);
        if(verbose){
          int i;
          for(i=0;i<16;i++){
            sprintf(md5text+2*i,"%02x",tmp->md5digest[i]);
            }
          }
        }
      if(verbose){
        printf("/%s%s - MD5 digest: %s\n",tmp->path->name,tmp->name,tmp->md5digest?md5text:"not calculated");
        }
      tmp=tmp->next;
      }while(tmp!=firstfile);
    if(status==1){        
      printf("Not enough memory.\n");
      }
    }
  return status;
  }

int find_duplicates(){

  struct fileinfo *tmp1=firstfile;
  struct fileinfo *tmp2=NULL;

  if(tmp1){ 
    printf("Finding duplicate files...\n");
    do{
      if(tmp1->linkto==NULL&&tmp1->md5digest!=NULL){
        tmp2=tmp1->next;
        while(tmp2!=firstfile){
          if(tmp1->lstat.st_size!=tmp2->lstat.st_size){
            break;
            }
          if(tmp2->linkto==NULL&&tmp2->md5digest!=NULL){
            if(!memcmp(tmp1->md5digest,tmp2->md5digest,sizeof(md5_byte_t)*16)){
              tmp2->linkto=tmp1;
              totalduplicatefiles++;
              totalduplicatebytes+=tmp2->lstat.st_size;
              if(verbose){
                printf("/%s%s same as /%s%s\n",tmp2->path->name,tmp2->name,tmp2->linkto->path->name,tmp2->linkto->name);
                }
              }
            }
          tmp2=tmp2->next;
          }
        }
      tmp1=tmp1->next;
      }while(tmp1!=firstfile);
    printf("Total %llu byte(s) in %lu duplicate file(s) (%u%%)\n",totalduplicatebytes,totalduplicatefiles,(totalbytes)?(int)(((double)totalduplicatebytes/(double)totalbytes)*(double)100):100);
    }
  return 0; 
  }  

int create_links(){

  int status=0;
  struct fileinfo *tmp=firstfile;
  char *linkname=NULL;
  char *linktoname=NULL;
  char *backupname=NULL;

  if(tmp){
    if(disableundo){
      printf("Removing duplicate files");
      }
    else{
      printf("Backuping duplicate files");
      }
    if(mode==1){
      printf(" and creating symbolic links");
      }
    printf("...\n");
    do{
      if(tmp->linkto){
        if(!(linkname=(char *)malloc(strlen(tmp->path->name)+strlen(tmp->name)+1))){
          status=1;
          break;
          }
        strcpy(linkname,tmp->path->name);
        strcat(linkname,tmp->name);
        if(!(backupname=(char *)malloc(strlen(tmp->path->name)+strlen(tmp->name)+11+1))){
          status=1;
          break;
          }
        strcpy(backupname,tmp->path->name);
        strcat(backupname,tmp->name);
        strcat(backupname,".dfl.backup");
        if(mode==1){
          if(!(linktoname=(char *)malloc(tmp->path->level*3+strlen(tmp->linkto->path->name)+strlen(tmp->linkto->name)+1))){
            status=1;
            break;
            }
          strcpy(linktoname,"");
          int idx;
          for(idx=0;idx<tmp->path->level;idx++){
            strcat(linktoname,"../");
            }
          strcat(linktoname,tmp->linkto->path->name);
          strcat(linktoname,tmp->linkto->name);
          }
        if(disableundo){
          if(remove(linkname)){
            printf("Can't remove file: %s\n",linkname);
            status=2;
            break;
            }
          }
        else{
          remove(backupname);
          if(rename(linkname,backupname)){
            printf("Can't create backup file: %s\n",backupname);
            status=2;
            break;
            }
          tmp->backup=1;
          }
        if(verbose && mode==2){
          printf("/%s\n",linkname);
          }
        if(mode==1){
          if(symlink(linktoname,linkname)){
            printf("Can't create symbolic link: %s -> %s\n",linkname,linktoname);
            status=2;
            break;
            }
          if(verbose && mode==1){
            printf("/%s -> %s\n",linkname,linktoname);
            }
          free(linktoname); 
          }
        free(backupname); 
        free(linkname);
        }
      tmp=tmp->next;
      }while(tmp!=firstfile);
    if(status==1){        
      printf("Not enough memory.\n");
      }
    }
  return status;
  }

int remove_backups(){

  int status=0;
  struct fileinfo *tmp=firstfile;
  char *backupname=NULL;

  if(tmp){ 
    printf("Removing backup files...\n");
    do{
      if(tmp->backup){
        if(!(backupname=(char *)malloc(strlen(tmp->path->name)+strlen(tmp->name)+11+1))){
          status=1;
          break;
          }
        strcpy(backupname,tmp->path->name);
        strcat(backupname,tmp->name);
        strcat(backupname,".dfl.backup");
        if(remove(backupname)){
          printf("Can't remove backup file: %s%s.dfl.backup\n",tmp->path->name,tmp->name);
          status=2;
          }
        else{
          if(verbose){
            printf("/%s%s.dfl.backup\n",tmp->path->name,tmp->name);
            }
          }
        free(backupname);
        }
      tmp=tmp->next;
      }while(tmp!=firstfile);
    if(status==1){        
      printf("Not enough memory.\n");
      }
    }
  return status;
  }

int undo(){

  int status=0;
  struct fileinfo *tmp=firstfile;
  char *filename=NULL;
  char *backupname=NULL;

  if(tmp){
    printf("Undo changes...\n");
    do{
      if(tmp->backup){
        if(!(filename=(char *)malloc(strlen(tmp->path->name)+strlen(tmp->name)+1))){
          status=1;
          break;
          }
        strcpy(filename,tmp->path->name);
        strcat(filename,tmp->name);
        if(!(backupname=(char *)malloc(strlen(tmp->path->name)+strlen(tmp->name)+11+1))){
          status=1;
          break;
          }
        strcpy(backupname,tmp->path->name);
        strcat(backupname,tmp->name);
        strcat(backupname,".dfl.backup");
        remove(filename);
        if(rename(backupname,filename)){
          printf("Can't restore file: %s%s\n",tmp->path->name,tmp->name);
          status=2;
          }
        else{
          if(verbose){
            printf("/%s%s\n",tmp->path->name,tmp->name);
            }
          }
        free(filename);
        free(backupname);
        }
      tmp=tmp->next;
      }while(tmp!=firstfile);
    if(status==1){        
      printf("Not enough memory.\n");
      }
    }
  return status;
  }
